@@ -1,23 +1,47 @@
-# Copyright 2010-2015 RethinkDB, all rights reserved.
+# Copyright 2018 RethinkDB
+#
+# Licensed under the Apache License, Version 2.0 (the 'License');
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an 'AS IS' BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+#
+# This file incorporates work covered by the following copyright:
+# Copyright 2010-2016 RethinkDB, all rights reserved.
+
 
 __all__ = [
-    "ReqlCursorEmpty", "RqlCursorEmpty",
-    "ReqlError", "RqlError",
-        "ReqlCompileError", "RqlCompileError",
-            "ReqlDriverCompileError",
-            "ReqlServerCompileError",
-        "ReqlRuntimeError", "RqlRuntimeError",
-            "ReqlQueryLogicError",
-                "ReqlNonExistenceError",
-            "ReqlResourceLimitError",
-            "ReqlUserError",
-            "ReqlInternalError",
-            "ReqlTimeoutError", "RqlTimeoutError",
-            "ReqlAvailabilityError",
-                "ReqlOpFailedError",
-                "ReqlOpIndeterminateError",
-        "ReqlDriverError", "RqlDriverError", "RqlClientError",
-            "ReqlAuthError"
+    "ReqlAuthError",
+    "ReqlAvailabilityError",
+    "ReqlCompileError",
+    "ReqlCursorEmpty",
+    "ReqlDriverCompileError",
+    "ReqlDriverError",
+    "ReqlError",
+    "ReqlInternalError",
+    "ReqlNonExistenceError",
+    "ReqlOpFailedError",
+    "ReqlOpIndeterminateError",
+    "ReqlPermissionError",
+    "ReqlQueryLogicError",
+    "ReqlResourceLimitError",
+    "ReqlRuntimeError",
+    "ReqlServerCompileError",
+    "ReqlTimeoutError",
+    "ReqlUserError",
+    "RqlClientError",
+    "RqlCompileError",
+    "RqlCursorEmpty",
+    "RqlDriverError",
+    "RqlError",
+    "RqlRuntimeError",
+    "RqlTimeoutError",
 ]
 
 import sys
@@ -26,7 +50,7 @@ try:
     unicode
 
     def convertForPrint(inputString):
-        if type(inputString) == unicode:
+        if isinstance(inputString, unicode):  # noqa: F821
             encoding = 'utf-8'
             if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
                 encoding = sys.stdout.encoding
@@ -39,9 +63,10 @@ except NameError:
 
 try:
     {}.iteritems
-    dict_items = lambda d: d.iteritems()
+
+    def dict_items(d): return d.iteritems()
 except AttributeError:
-    dict_items = lambda d: d.items()
+    def dict_items(d): return d.items()
 
 
 class ReqlCursorEmpty(Exception):
@@ -49,6 +74,7 @@ class ReqlCursorEmpty(Exception):
         super(ReqlCursorEmpty, self).__init__("Cursor is empty.")
         self.message = "Cursor is empty."
 
+
 RqlCursorEmpty = ReqlCursorEmpty
 
 
@@ -69,17 +95,24 @@ class ReqlError(Exception):
                 self.query_printer.print_query(),
                 self.query_printer.print_carrots()))
 
+    def __repr__(self):
+        return "<%s instance: %s >" % (self.__class__.__name__, str(self))
+
+
 RqlError = ReqlError
 
 
 class ReqlCompileError(ReqlError):
     pass
 
+
 RqlCompileError = ReqlCompileError
 
+
 class ReqlDriverCompileError(ReqlCompileError):
     pass
 
+
 class ReqlServerCompileError(ReqlCompileError):
     pass
 
@@ -87,6 +120,7 @@ class ReqlServerCompileError(ReqlCompileError):
 class ReqlRuntimeError(ReqlError):
     pass
 
+
 RqlRuntimeError = ReqlRuntimeError
 
 
@@ -122,46 +156,47 @@ class ReqlOpIndeterminateError(ReqlAvailabilityError):
     pass
 
 
+class ReqlPermissionError(ReqlRuntimeError):
+    pass
+
+
 class ReqlDriverError(ReqlError):
     pass
 
+
 RqlClientError = ReqlDriverError
 RqlDriverError = ReqlDriverError
 
+
 class ReqlAuthError(ReqlDriverError):
+    def __init__(self, msg, host=None, port=None):
+        if host is not None and port is not None:
+            msg = "Could not connect to {}:{}, {}".format(host, port, msg)
+        super(ReqlAuthError, self).__init__(msg)
+
+
+class _ReqlTimeoutError(ReqlDriverError):
     def __init__(self, host=None, port=None):
-        if host is None or port is None:
-            super(ReqlDriverError, self).__init__("Incorrect authentication key.")
-        else:
-            super(ReqlDriverError, self).__init__(
-                "Could not connect to %s:%d, incorrect authentication key." %
-                (host, port))
+        msg = "Operation timed out."
+        if host is not None and port is not None:
+            msg = "Could not connect to {}:{}, {}".format(host, port, msg)
+        super(_ReqlTimeoutError, self).__init__(msg)
 
 
 try:
-    class ReqlTimeoutError(ReqlDriverError, TimeoutError):
-        def __init__(self, host=None, port=None):
-            if host is None or port is None:
-                super(ReqlDriverError, self).__init__("Operation timed out.")
-            else:
-                super(ReqlDriverError, self).__init__(
-                    "Could not connect to %s:%d, operation timed out." % (host, port))
+    class ReqlTimeoutError(_ReqlTimeoutError, TimeoutError):
+        pass
 except NameError:
-    class ReqlTimeoutError(ReqlDriverError):
-        def __init__(self, host=None, port=None):
-            if host is None or port is None:
-                super(ReqlDriverError, self).__init__("Operation timed out.")
-            else:
-                super(ReqlDriverError, self).__init__(
-                    "Could not connect to %s:%d, operation timed out." % (host, port))
+    class ReqlTimeoutError(_ReqlTimeoutError):
+        pass
 
 RqlTimeoutError = ReqlTimeoutError
 
 
 class QueryPrinter(object):
-    def __init__(self, root, frames=[]):
+    def __init__(self, root, frames=None):
         self.root = root
-        self.frames = frames
+        self.frames = list(frames or ())
 
     def print_query(self):
         return ''.join(self.compose_term(self.root))
@@ -170,7 +205,7 @@ class QueryPrinter(object):
         return ''.join(self.compose_carrots(self.root, self.frames))
 
     def compose_term(self, term):
-        args = [self.compose_term(a) for a in term.args]
+        args = [self.compose_term(a) for a in term._args]
         optargs = {}
         for k, v in dict_items(term.optargs):
             optargs[k] = self.compose_term(v)
@@ -184,7 +219,7 @@ class QueryPrinter(object):
         cur_frame = frames[0]
         args = [self.compose_carrots(arg, frames[1:])
                 if cur_frame == i else self.compose_term(arg)
-                for i, arg in enumerate(term.args)]
+                for i, arg in enumerate(term._args)]
 
         optargs = {}
         for k, v in dict_items(term.optargs):
@@ -208,10 +243,16 @@ class T(object):
 
     def __iter__(self):
         itr = iter(self.seq)
-        for sub in next(itr):
-            yield sub
+
+        try:
+            for sub in next(itr):
+                yield sub
+        except StopIteration:
+            return
+
         for token in itr:
             for sub in self.intsp:
                 yield sub
+
             for sub in token:
                 yield sub
