@@ -1,57 +1,85 @@
-# Copyright 2010-2015 RethinkDB, all rights reserved.
+# Copyright 2018 RethinkDB
+#
+# Licensed under the Apache License, Version 2.0 (the 'License');
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an 'AS IS' BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+#
+# This file incorporates work covered by the following copyright:
+# Copyright 2010-2016 RethinkDB, all rights reserved.
+
 
 import collections
 import errno
-import imp
 import numbers
-import os
+import pprint
 import socket
 import ssl
 import struct
 import time
 
-from . import ql2_pb2 as p
+try:
+    from urllib.parse import urlparse, parse_qs
+except ImportError:
+    from urlparse import urlparse, parse_qs
 
-__all__ = ['connect', 'set_loop_type', 'Connection', 'Cursor']
+from rethinkdb import ql2_pb2
+from rethinkdb.ast import DB, ReQLDecoder, ReQLEncoder, Repl, expr
+from rethinkdb.errors import (
+    ReqlAuthError,
+    ReqlCursorEmpty,
+    ReqlDriverError,
+    ReqlError,
+    ReqlInternalError,
+    ReqlNonExistenceError,
+    ReqlOpFailedError,
+    ReqlOpIndeterminateError,
+    ReqlPermissionError,
+    ReqlQueryLogicError,
+    ReqlResourceLimitError,
+    ReqlRuntimeError,
+    ReqlServerCompileError,
+    ReqlTimeoutError,
+    ReqlUserError)
+from rethinkdb.handshake import HandshakeV1_0
+from rethinkdb.logger import default_logger
 
-pErrorType = p.Response.ErrorType
-pResponse = p.Response.ResponseType
-pQuery = p.Query.QueryType
+__all__ = ['Connection', 'Cursor', 'DEFAULT_PORT', 'DefaultConnection', 'make_connection']
+
+
+DEFAULT_PORT = 28015
+
+pErrorType = ql2_pb2.Response.ErrorType
+pResponse = ql2_pb2.Response.ResponseType
+pQuery = ql2_pb2.Query.QueryType
 
-from .errors import *
-from .ast import RqlQuery, RqlTopLevelQuery, DB, Repl, ReQLDecoder, ReQLEncoder
 
 try:
     from ssl import match_hostname, CertificateError
 except ImportError:
-    from backports.ssl_match_hostname import match_hostname, CertificateError
-
-try:
-    xrange
-except NameError:
-    xrange = range
+    from .backports.ssl_match_hostname import match_hostname, CertificateError
 
 try:
     {}.iteritems
-    dict_items = lambda d: d.iteritems()
-except AttributeError:
-    dict_items = lambda d: d.items()
 
-def decodeUTF(inputPipe):
-    # attempt to decode input as utf-8 with fallbacks to get something useful
-    try:
-        return inputPipe.decode('utf-8', errors='ignore')
-    except TypeError:
-        try:
-            return inputPipe.decode('utf-8')
-        except UnicodeError:
-            return repr(inputPipe)
+    def dict_items(d): return d.iteritems()
+except AttributeError:
+    def dict_items(d): return d.items()
+
 
 def maybe_profile(value, res):
     if res.profile is not None:
         return {'value': value, 'profile': res.profile}
     return value
 
+
 class Query(object):
     def __init__(self, type, token, term, global_optargs):
         self.type = type
@@ -59,7 +87,7 @@ class Query(object):
         self.term = term
         self.global_optargs = global_optargs
 
-        global_optargs = global_optargs or { }
+        global_optargs = global_optargs or {}
         self._json_encoder = global_optargs.pop('json_encoder', None)
         self._json_decoder = global_optargs.pop('json_decoder', None)
 
@@ -68,7 +96,7 @@ class Query(object):
         if self.term is not None:
             message.append(self.term)
         if self.global_optargs is not None:
-            message.append(self.global_optargs)
+            message.append(expr(self.global_optargs))
         query_str = reql_encoder.encode(message).encode('utf-8')
         query_header = struct.pack('<QL', self.token, len(query_str))
         return query_header + query_str
@@ -101,7 +129,8 @@ class Response(object):
                 pErrorType.NON_EXISTENCE: ReqlNonExistenceError,
                 pErrorType.OP_FAILED: ReqlOpFailedError,
                 pErrorType.OP_INDETERMINATE: ReqlOpIndeterminateError,
-                pErrorType.USER: ReqlUserError
+                pErrorType.USER: ReqlUserError,
+                pErrorType.PERMISSION_ERROR: ReqlPermissionError
             }.get(self.error_type, ReqlRuntimeError)(
                 self.data[0], query.term, self.backtrace)
         return ReqlDriverError(("Unknown Response type %d encountered" +
@@ -130,6 +159,8 @@ class Response(object):
 #     Exception - an error has occurred in the cursor and should be raised
 #         to the user once all results in `items` have been returned.  This
 #         will be a ReqlCursorEmpty exception if the cursor completed successfully.
+#         TODO @gabor-boros: We should not set the `errors` to ReqlCursorEmpty, due
+#         to it is not an error but a success state.
 #
 # A class that derives from this should implement the following functions:
 #     def _get_next(self, timeout):
@@ -138,7 +169,7 @@ class Response(object):
 #     def _empty_error(self):
 #         which returns the appropriate error to be raised when the cursor is empty
 class Cursor(object):
-    def __init__(self, conn_instance, query, first_response, items_type = collections.deque):
+    def __init__(self, conn_instance, query, first_response, items_type=collections.deque):
         self.conn = conn_instance
         self.query = query
         self.items = items_type()
@@ -152,6 +183,12 @@ class Cursor(object):
         self._maybe_fetch_batch()
         self._extend_internal(first_response)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        self.close()
+
     def close(self):
         if self.error is None:
             self.error = self._empty_error()
@@ -193,18 +230,43 @@ class Cursor(object):
             del self.conn._cursor_cache[res.token]
 
     def __str__(self):
-        val_str = ', '.join(map(repr, self.items[:10]))
-        if len(self.items) > 10 or self.error is None:
-            val_str += ', ...'
-
+        val_str = pprint.pformat([self.items[x] for x in range(min(10, len(self.items)))] +
+                                 (['...'] if len(self.items) > 10 else []))
+        if val_str.endswith("'...']"):
+            val_str = val_str[:-len("'...']")] + "...]"
+        spacer_str = '\n' if '\n' in val_str else ''
         if self.error is None:
-            err_str = 'streaming'
+            status_str = 'streaming'
         elif isinstance(self.error, ReqlCursorEmpty):
-            err_str = 'done streaming'
+            status_str = 'done streaming'
         else:
-            err_str = 'error: %s' % repr(self.error)
+            status_str = 'error: %s' % str(self.error)
 
-        return "%s (%s):\n[%s]" % (object.__str__(self), err_str, val_str)
+        return "%s.%s (%s): %s%s" % (
+            self.__class__.__module__, self.__class__.__name__,
+            status_str,
+            spacer_str, val_str
+        )
+
+    def __repr__(self):
+        val_str = pprint.pformat([self.items[x] for x in range(min(10, len(self.items)))] +
+                                 (['...'] if len(self.items) > 10 else []))
+        if val_str.endswith("'...']"):
+            val_str = val_str[:-len("'...']")] + "...]"
+        spacer_str = '\n' if '\n' in val_str else ''
+        if self.error is None:
+            status_str = 'streaming'
+        elif isinstance(self.error, ReqlCursorEmpty):
+            status_str = 'done streaming'
+        else:
+            status_str = 'error: %s' % repr(self.error)
+
+        return "<%s.%s object at %s (%s): %s%s>" % (
+            self.__class__.__module__, self.__class__.__name__,
+            hex(id(self)),
+            status_str,
+            spacer_str, val_str
+        )
 
     def _error(self, message):
         # Set an error and extend with a dummy response to trigger any waiters
@@ -262,31 +324,64 @@ class SocketWrapper(object):
             self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
 
             if len(self.ssl) > 0:
-                ssl_context = self._get_ssl_context(self.ssl["ca_certs"])
                 try:
-                    self._socket = ssl_context.wrap_socket(self._socket,
-                                                           server_hostname=self.host)
-                except IOError as exc:
+                    if hasattr(ssl, 'SSLContext'):  # Python2.7 and 3.2+, or backports.ssl
+                        ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
+                        if hasattr(ssl_context, "options"):
+                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv2", 0)
+                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv3", 0)
+                        ssl_context.verify_mode = ssl.CERT_REQUIRED
+                        ssl_context.check_hostname = True  # redundant with match_hostname
+                        ssl_context.load_verify_locations(self.ssl["ca_certs"])
+                        self._socket = ssl_context.wrap_socket(self._socket, server_hostname=self.host)
+                    else:  # this does not disable SSLv2 or SSLv3
+                        self._socket = ssl.wrap_socket(
+                            self._socket, cert_reqs=ssl.CERT_REQUIRED, ssl_version=ssl.PROTOCOL_SSLv23,
+                            ca_certs=self.ssl["ca_certs"])
+                except IOError as err:
                     self._socket.close()
-                    raise ReqlDriverError("SSL handshake failed: %s" % (str(exc),))
+
+                    if 'EOF occurred in violation of protocol' in str(
+                            err) or 'sslv3 alert handshake failure' in str(err):
+                        # probably on an older version of OpenSSL
+                        raise ReqlDriverError(
+                            "SSL handshake failed, likely because Python is linked against an old version of OpenSSL "
+                            "that does not support either TLSv1.2 or any of the allowed ciphers. This can be worked "
+                            "around by lowering the security setting on the server with the options "
+                            "`--tls-min-protocol TLSv1 --tls-ciphers "
+                            "EECDH+AESGCM:EDH+AESGCM:AES256+EECDH:AES256+EDH:AES256-SHA` (see server log for more "
+                            "information): %s" % str(err)
+                        )
+                    else:
+                        raise ReqlDriverError(
+                            "SSL handshake failed (see server log for more information): %s" %
+                            str(err))
                 try:
                     match_hostname(self._socket.getpeercert(), hostname=self.host)
                 except CertificateError:
                     self._socket.close()
                     raise
 
-            self.sendall(parent._parent.handshake)
-
-            # The response from the server is a null-terminated string
-            response = b''
+            parent._parent.handshake.reset()
+            response = None
             while True:
-                char = self.recvall(1, deadline)
-                if char == b'\0':
+                request = parent._parent.handshake.next_message(response)
+                if request is None:
                     break
-                response += char
-        except ReqlAuthError:
-            raise
-        except ReqlTimeoutError:
+                # This may happen in the `V1_0` protocol where we send two requests as
+                # an optimization, then need to read each separately
+                if request is not "":
+                    self.sendall(request)
+
+                # The response from the server is a null-terminated string
+                response = b''
+                while True:
+                    char = self.recvall(1, deadline)
+                    if char == b'\0':
+                        break
+                    response += char
+        except (ReqlAuthError, ReqlTimeoutError):
+            self.close()
             raise
         except ReqlDriverError as ex:
             self.close()
@@ -300,27 +395,20 @@ class SocketWrapper(object):
         except Exception as ex:
             self.close()
             raise ReqlDriverError("Could not connect to %s:%s. Error: %s" %
-                                  (self.host, self.port, ex))
-
-        if response != b"SUCCESS":
-            self.close()
-            message = decodeUTF(response).strip()
-            if message == "ERROR: Incorrect authorization key.":
-                raise ReqlAuthError(self.host, self.port)
-            else:
-                raise ReqlDriverError("Server dropped connection with message: \"%s\"" %
-                                      (message, ))
+                                  (self.host, self.port, str(ex)))
 
     def is_open(self):
         return self._socket is not None
 
     def close(self):
-        if self._socket is not None:
+        if self.is_open():
             try:
                 self._socket.shutdown(socket.SHUT_RDWR)
                 self._socket.close()
-            except Exception:
-                pass
+            except ReqlError as exc:
+                default_logger.error(exc.message)
+            except Exception as exc:
+                default_logger.error(exc)
             finally:
                 self._socket = None
 
@@ -343,20 +431,18 @@ class SocketWrapper(object):
                         self.close()
                         raise ReqlDriverError("Connection is closed.")
                     elif ex.errno == errno.EWOULDBLOCK:
+                        self.close()
                         # This should only happen with a timeout of 0
                         raise ReqlTimeoutError(self.host, self.port)
                     elif ex.errno != errno.EINTR:
-                        self.close()
                         raise ReqlDriverError(('Connection interrupted ' +
-                                              'receiving from %s:%s - %s') %
-                                             (self.host, self.port, str(ex)))
+                                               'receiving from %s:%s - %s') %
+                                              (self.host, self.port, str(ex)))
                 except Exception as ex:
                     self.close()
                     raise ReqlDriverError('Error receiving from %s:%s - %s' %
-                                         (self.host, self.port, str(ex)))
-                except:
-                    self.close()
-                    raise
+                                          (self.host, self.port, str(ex)))
+
             if len(chunk) == 0:
                 self.close()
                 raise ReqlDriverError("Connection is closed.")
@@ -375,27 +461,16 @@ class SocketWrapper(object):
                 elif ex.errno != errno.EINTR:
                     self.close()
                     raise ReqlDriverError(('Connection interrupted ' +
-                                          'sending to %s:%s - %s') %
-                                         (self.host, self.port, str(ex)))
+                                           'sending to %s:%s - %s') %
+                                          (self.host, self.port, str(ex)))
             except Exception as ex:
                 self.close()
                 raise ReqlDriverError('Error sending to %s:%s - %s' %
-                                     (self.host, self.port, str(ex)))
-            except:
+                                      (self.host, self.port, str(ex)))
+            except BaseException:
                 self.close()
                 raise
 
-    def _get_ssl_context(self, ca_certs):
-        ctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-        if hasattr(ctx, "options"):
-            ctx.options |= getattr(ssl, "OP_NO_SSLv2", 0)
-            ctx.options |= getattr(ssl, "OP_NO_SSLv3", 0)
-
-        ctx.verify_mode = ssl.CERT_REQUIRED
-        ctx.check_hostname = True
-        ctx.load_verify_locations(ca_certs)
-        return ctx
-
 
 class ConnectionInstance(object):
     def __init__(self, parent):
@@ -405,6 +480,14 @@ class ConnectionInstance(object):
         self._socket = None
         self._closing = False
 
+    def client_port(self):
+        if self.is_open():
+            return self._socket._socket.getsockname()[1]
+
+    def client_address(self):
+        if self.is_open():
+            return self._socket._socket.getsockname()[0]
+
     def connect(self, timeout):
         self._socket = SocketWrapper(self, timeout)
         return self._parent
@@ -412,7 +495,7 @@ class ConnectionInstance(object):
     def is_open(self):
         return self._socket.is_open()
 
-    def close(self, noreply_wait, token):
+    def close(self, noreply_wait=False, token=None):
         self._closing = True
 
         # Cursors may remove themselves when errored, so copy a list of them
@@ -485,7 +568,7 @@ class ConnectionInstance(object):
                     self._parent._get_json_decoder(query))
             elif not self._closing:
                 # This response is corrupted or not intended for us
-                self.close(False, None)
+                self.close()
                 raise ReqlDriverError("Unexpected response received.")
 
 
@@ -494,16 +577,15 @@ class Connection(object):
     _json_decoder = ReQLDecoder
     _json_encoder = ReQLEncoder
 
-    def __init__(self, conn_type, host, port, db, auth_key, timeout, ssl, **kwargs):
+    def __init__(self, conn_type, host, port, db, auth_key, user, password, timeout, ssl, _handshake_version, **kwargs):
         self.db = db
-        self.auth_key = auth_key.encode('ascii')
-        self.handshake = \
-            struct.pack("<2L", p.VersionDummy.Version.V0_4, len(self.auth_key)) + \
-            self.auth_key + \
-            struct.pack("<L", p.VersionDummy.Protocol.JSON)
 
         self.host = host
-        self.port = port
+        try:
+            self.port = int(port)
+        except ValueError:
+            raise ReqlDriverError("Could not convert port %r to an integer." % port)
+
         self.connect_timeout = timeout
 
         self.ssl = ssl
@@ -518,19 +600,36 @@ class Connection(object):
         if 'json_decoder' in kwargs:
             self._json_decoder = kwargs.pop('json_decoder')
 
+        if auth_key is None and password is None:
+            auth_key = password = ''
+        elif auth_key is None and password is not None:
+            auth_key = password
+        elif auth_key is not None and password is None:
+            password = auth_key
+        else:
+            # auth_key is not None and password is not None
+            raise ReqlDriverError("`auth_key` and `password` are both set.")
+
+        if _handshake_version == 4:
+            raise NotImplementedError("The v0.4 handshake was removed.")
+
+        self.handshake = HandshakeV1_0(
+            self._json_decoder(), self._json_encoder(), self.host, self.port, user, password)
+
+    def client_port(self):
+        if self.is_open():
+            return self._instance.client_port()
+
+    def client_address(self):
+        if self.is_open():
+            return self._instance.client_address()
+
     def reconnect(self, noreply_wait=True, timeout=None):
         if timeout is None:
             timeout = self.connect_timeout
 
         self.close(noreply_wait)
 
-        # Do this here rather than in the constructor so that we don't throw
-        # in the constructor (which doesn't play well with Tornado)
-        try:
-            self.port = int(self.port)
-        except ValueError:
-            raise ReqlDriverError("Could not convert port %s to an integer." % self.port)
-
         self._instance = self._conn_type(self, **self._child_kwargs)
         return self._instance.connect(timeout)
 
@@ -603,31 +702,56 @@ class Connection(object):
     def _get_json_encoder(self, query):
         return (query._json_encoder or self._json_encoder)()
 
+
 class DefaultConnection(Connection):
     def __init__(self, *args, **kwargs):
         Connection.__init__(self, ConnectionInstance, *args, **kwargs)
 
-connection_type = DefaultConnection
 
-def connect(host='localhost', port=28015, db=None, auth_key="", timeout=20, ssl=dict(), **kwargs):
-    conn = connection_type(host, port, db, auth_key, timeout, ssl, **kwargs)
+def make_connection(
+        connection_type,
+        host=None,
+        port=None,
+        db=None,
+        auth_key=None,
+        user=None,
+        password=None,
+        timeout=20,
+        ssl=None,
+        url=None,
+        _handshake_version=10,
+        **kwargs):
+    if url:
+        connection_string = urlparse(url)
+        query_string = parse_qs(connection_string.query)
+
+        user = connection_string.username
+        password = connection_string.password
+        host = connection_string.hostname
+        port = connection_string.port
+
+        db = connection_string.path.replace("/", "") or None
+        auth_key = query_string.get("auth_key")
+        timeout = query_string.get("timeout")
+
+        if auth_key:
+            auth_key = auth_key[0]
+
+        if timeout:
+            timeout = int(timeout[0])
+
+
+    host = host or 'localhost'
+    port = port or DEFAULT_PORT
+    user = user or 'admin'
+    timeout = timeout or 20
+    ssl = ssl or dict()
+    _handshake_version = _handshake_version or 10
+
+    # The internal APIs will wait for none to deal with auth_key and password
+    # TODO: refactor when we drop python2
+    if not password and not password is None:
+        password = None
+
+    conn = connection_type(host, port, db, auth_key, user, password, timeout, ssl, _handshake_version, **kwargs)
     return conn.reconnect(timeout=timeout)
-
-def set_loop_type(library):
-    global connection_type
-    
-    # find module file
-    moduleName = 'net_%s' % library
-    modulePath = None
-    driverDir = os.path.realpath(os.path.dirname(__file__))
-    if os.path.isfile(os.path.join(driverDir, library + '_net', moduleName + '.py')):
-        modulePath = os.path.join(driverDir, library + '_net', moduleName + '.py')
-    else:
-        raise ValueError('Unknown loop type: %r' % library)
-    
-    # load the module
-    moduleFile, pathName, desc = imp.find_module(moduleName, [os.path.dirname(modulePath)])
-    module = imp.load_module('rethinkdb.' + moduleName, moduleFile, pathName, desc)
-    
-    # set the connection type
-    connection_type = module.Connection
